@@ -32,6 +32,10 @@
 #include <stdlib.h>
 #include <unistd.h>
 
+#if __linux__
+#  include <sys/syscall.h>
+#endif
+
 // NetBSD opens a file descriptor for each sem_init()
 #if defined(_POSIX_SEMAPHORES) && !__NetBSD__
 #  define BFS_POSIX_SEMAPHORES _POSIX_SEMAPHORES
@@ -245,7 +249,69 @@ static void *rcu_update(struct rcu *rcu, void *ptr) {
 	return rcu_decode(arc_wait(prev));
 }
 
+/**
+ * An RCU-protected linked list.
+ */
+struct rcu_list {
+	/** The first node in the list. */
+	struct rcu head;
+	/** &last->next */
+	struct rcu *tail;
+};
+
+/**
+ * An rcu_list node.
+ */
+struct rcu_node {
+	/** The RCU pointer to this node. */
+	struct rcu *self;
+	/** The next node in the list. */
+	struct rcu next;
+};
+
+/** Initialize an rcu_list. */
+static void rcu_list_init(struct rcu_list *list) {
+	rcu_init(&list->head, NULL);
+	list->tail = &list->head;
+}
+
+/** Append to an rcu_list. */
+static void rcu_list_append(struct rcu_list *list, struct rcu_node *node) {
+	node->self = list->tail;
+	list->tail = &node->next;
+	rcu_init(&node->next, NULL);
+	rcu_update(node->self, node);
+}
+
+/** Remove from an rcu_list. */
+static void rcu_list_remove(struct rcu_list *list, struct rcu_node *node) {
+	struct rcu_node *next = rcu_peek(&node->next);
+	rcu_update(node->self, next);
+	if (next) {
+		next->self = node->self;
+	} else {
+		list->tail = &list->head;
+	}
+	rcu_destroy(&node->next);
+}
+
+/** Iterate over an rcu_list. */
+#define for_rcu(type, node, list) \
+	for_rcu_(type, node, (list), node##_slot_, node##_prev_, node##_done_)
+
+#define for_rcu_(type, node, list, slot, prev, done) \
+	/* This outer loop is just for declaring variables; it iterates once. */ \
+	for (struct arc *slot, *prev, **done = NULL; !done && (done = &slot); ) \
+		for (type *node = rcu_read(&list->head, &slot); \
+		     node || (arc_put(slot), false); \
+		     (prev = slot, \
+		      node = rcu_read(&((struct rcu_node *)node)->next, &slot), \
+		      arc_put(prev)))
+
 struct sighook {
+	/** The RCU list node (must be the first field). */
+	struct rcu_node node;
+
 	/** The signal being hooked, or 0 for atsigexit(). */
 	int sig;
 	/** Signal hook flags. */
@@ -254,53 +320,15 @@ struct sighook {
 	sighook_fn *fn;
 	/** An argument to pass to the function. */
 	void *arg;
-
-	/** The RCU pointer to this hook. */
-	struct rcu *self;
-	/** The next hook in the list. */
-	struct rcu next;
+	/** Flag for SH_ONESHOT. */
+	atomic bool armed;
 };
 
-/**
- * An RCU-protected linked list of signal hooks.
- */
-struct siglist {
-	/** The first hook in the list. */
-	struct rcu head;
-	/** &last->next */
-	struct rcu *tail;
-};
-
-/** Initialize a siglist. */
-static void siglist_init(struct siglist *list) {
-	rcu_init(&list->head, NULL);
-	list->tail = &list->head;
-}
-
-/** Append a hook to a linked list. */
-static void sigpush(struct siglist *list, struct sighook *hook) {
-	hook->self = list->tail;
-	list->tail = &hook->next;
-	rcu_init(&hook->next, NULL);
-	rcu_update(hook->self, hook);
-}
-
-/** Remove a hook from the linked list. */
-static void sigpop(struct siglist *list, struct sighook *hook) {
-	struct sighook *next = rcu_peek(&hook->next);
-	rcu_update(hook->self, next);
-	if (next) {
-		next->self = hook->self;
-	} else {
-		list->tail = &list->head;
-	}
-}
-
 /** The lists of signal hooks. */
-static struct siglist sighooks[64];
+static struct rcu_list sighooks[64];
 
 /** Get the hook list for a particular signal. */
-static struct siglist *siglist(int sig) {
+static struct rcu_list *siglist(int sig) {
 	return &sighooks[sig % countof(sighooks)];
 }
 
@@ -336,22 +364,31 @@ static const int FATAL_SIGNALS[] = {
 	SIGHUP,
 	SIGILL,
 	SIGINT,
+#ifdef SIGIO
+	SIGIO,
+#endif
 	SIGPIPE,
-	SIGQUIT,
-	SIGSEGV,
-	SIGTERM,
-	SIGUSR1,
-	SIGUSR2,
 #ifdef SIGPOLL
 	SIGPOLL,
 #endif
 #ifdef SIGPROF
 	SIGPROF,
 #endif
+#ifdef SIGPWR
+	SIGPWR,
+#endif
+	SIGQUIT,
+	SIGSEGV,
+#ifdef SIGSTKFLT
+	SIGSTKFLT,
+#endif
 #ifdef SIGSYS
 	SIGSYS,
 #endif
+	SIGTERM,
 	SIGTRAP,
+	SIGUSR1,
+	SIGUSR2,
 #ifdef SIGVTALRM
 	SIGVTALRM,
 #endif
@@ -383,7 +420,9 @@ static bool is_fatal(int sig) {
 
 /** Reraise a fatal signal. */
 _noreturn
-static void reraise(int sig) {
+static void reraise(siginfo_t *info) {
+	int sig = info->si_signo;
+
 	// Restore the default signal action
 	if (signal(sig, SIG_DFL) == SIG_ERR) {
 		goto fail;
@@ -397,42 +436,70 @@ static void reraise(int sig) {
 		goto fail;
 	}
 
+#if __linux__
+	// On Linux, try to re-raise the exact siginfo_t (since 3.9, a process can
+	// signal itself with any siginfo_t)
+	pid_t tid = syscall(SYS_gettid);
+	syscall(SYS_rt_tgsigqueueinfo, getpid(), tid, sig, info);
+#endif
+
 	raise(sig);
 fail:
 	abort();
 }
 
+/** Check whether we should run a hook. */
+static bool should_run(int sig, struct sighook *hook) {
+	if (hook->sig != sig && hook->sig != 0) {
+		return false;
+	}
+
+	if (hook->flags & SH_ONESHOT) {
+		if (!exchange(&hook->armed, false, relaxed)) {
+			return false;
+		}
+	}
+
+	return true;
+}
+
 /** Find any matching hooks and run them. */
-static enum sigflags run_hooks(struct siglist *list, int sig, siginfo_t *info) {
+static enum sigflags run_hooks(struct rcu_list *list, int sig, siginfo_t *info) {
 	enum sigflags ret = 0;
 
-	struct arc *slot = NULL;
-	struct sighook *hook = rcu_read(&list->head, &slot);
-	while (hook) {
-		if (hook->sig == sig || hook->sig == 0) {
+	for_rcu (struct sighook, hook, list) {
+		if (should_run(sig, hook)) {
 			hook->fn(sig, info, hook->arg);
 			ret |= hook->flags;
 		}
-
-		struct arc *prev = slot;
-		hook = rcu_read(&hook->next, &slot);
-		arc_put(prev);
 	}
 
-	arc_put(slot);
 	return ret;
 }
 
 /** Dispatches a signal to the registered handlers. */
 static void sigdispatch(int sig, siginfo_t *info, void *context) {
-	// https://pubs.opengroup.org/onlinepubs/9799919799/functions/V2_chap02.html#tag_16_04_03_03
+	// If we get a fault (e.g. a "real" SIGSEGV, not something like
+	// kill(..., SIGSEGV)), don't try to run signal hooks, since we could be
+	// in an arbitrarily corrupted state.
 	//
-	//     The behavior of a process is undefined after it returns normally
-	//     from a signal-catching function for a SIGBUS, SIGFPE, SIGILL, or
-	//     SIGSEGV signal that was not generated by kill(), sigqueue(), or
-	//     raise().
+	// POSIX says that returning normally from a signal handler for a fault
+	// is undefined.  But in practice, it's better to uninstall the handler
+	// and return, which will re-run the faulting instruction and cause us
+	// to die "correctly" (e.g. with a core dump pointing at the faulting
+	// instruction, not reraise()).
 	if (is_fault(info)) {
-		reraise(sig);
+		// On macOS, we cannot reliably distinguish between faults and
+		// asynchronous signals.  For example, pkill -SEGV bfs will
+		// result in si_code == SEGV_ACCERR.  So we always re-raise the
+		// signal, because just returning would cause us to ignore
+		// asynchronous SIG{BUS,ILL,SEGV}.
+#if !__APPLE__
+		if (signal(sig, SIG_DFL) != SIG_ERR) {
+			return;
+		}
+#endif
+		reraise(info);
 	}
 
 	// https://pubs.opengroup.org/onlinepubs/9799919799/functions/V2_chap02.html#tag_16_04_04
@@ -445,40 +512,58 @@ static void sigdispatch(int sig, siginfo_t *info, void *context) {
 	int error = errno;
 
 	// Run the normal hooks
-	struct siglist *list = siglist(sig);
+	struct rcu_list *list = siglist(sig);
 	enum sigflags flags = run_hooks(list, sig, info);
 
 	// Run the atsigexit() hooks, if we're exiting
 	if (!(flags & SH_CONTINUE) && is_fatal(sig)) {
 		list = siglist(0);
 		run_hooks(list, sig, info);
-		reraise(sig);
+		reraise(info);
 	}
 
 	errno = error;
 }
 
+/** A saved signal handler, for sigreset() to restore. */
+struct sigsave {
+	struct rcu_node node;
+	int sig;
+	struct sigaction action;
+};
+
+/** The list of saved signal handlers. */
+static struct rcu_list saved;
+/** `saved` initialization status (since rcu_list_init() isn't atomic). */
+static atomic bool initialized = false;
+
 /** Make sure our signal handler is installed for a given signal. */
 static int siginit(int sig) {
+#ifdef SA_RESTART
+#  define BFS_SA_RESTART SA_RESTART
+#else
+#  define BFS_SA_RESTART 0
+#endif
+
 	static struct sigaction action = {
 		.sa_sigaction = sigdispatch,
-		.sa_flags = SA_RESTART | SA_SIGINFO,
+		.sa_flags = BFS_SA_RESTART | SA_SIGINFO,
 	};
 
 	static sigset_t signals;
-	static bool initialized = false;
 
-	if (!initialized) {
+	if (!load(&initialized, relaxed)) {
 		if (sigemptyset(&signals) != 0
 		    || sigemptyset(&action.sa_mask) != 0) {
 			return -1;
 		}
 
 		for (size_t i = 0; i < countof(sighooks); ++i) {
-			siglist_init(&sighooks[i]);
+			rcu_list_init(&sighooks[i]);
 		}
 
-		initialized = true;
+		rcu_list_init(&saved);
+		store(&initialized, true, release);
 	}
 
 	int installed = sigismember(&signals, sig);
@@ -488,14 +573,32 @@ static int siginit(int sig) {
 		return 0;
 	}
 
+	sigset_t updated = signals;
+	if (sigaddset(&updated, sig) != 0) {
+		return -1;
+	}
+
+	struct sigaction original;
+	if (sigaction(sig, NULL, &original) != 0) {
+		return -1;
+	}
+
+	struct sigsave *save = ALLOC(struct sigsave);
+	if (!save) {
+		return -1;
+	}
+
+	save->sig = sig;
+	save->action = original;
+	rcu_list_append(&saved, &save->node);
+
 	if (sigaction(sig, &action, NULL) != 0) {
+		rcu_list_remove(&saved, &save->node);
+		free(save);
 		return -1;
 	}
 
-	if (sigaddset(&signals, sig) != 0) {
-		return -1;
-	}
-
+	signals = updated;
 	return 0;
 }
 
@@ -510,9 +613,10 @@ static struct sighook *sighook_impl(int sig, sighook_fn *fn, void *arg, enum sig
 	hook->flags = flags;
 	hook->fn = fn;
 	hook->arg = arg;
+	atomic_init(&hook->armed, true);
 
-	struct siglist *list = siglist(sig);
-	sigpush(list, hook);
+	struct rcu_list *list = siglist(sig);
+	rcu_list_append(list, &hook->node);
 	return hook;
 }
 
@@ -558,11 +662,24 @@ void sigunhook(struct sighook *hook) {
 
 	mutex_lock(&sigmutex);
 
-	struct siglist *list = siglist(hook->sig);
-	sigpop(list, hook);
+	struct rcu_list *list = siglist(hook->sig);
+	rcu_list_remove(list, &hook->node);
 
 	mutex_unlock(&sigmutex);
 
-	rcu_destroy(&hook->next);
 	free(hook);
 }
+
+int sigreset(void) {
+	if (!load(&initialized, acquire)) {
+		return 0;
+	}
+
+	for_rcu (struct sigsave, save, &saved) {
+		if (sigaction(save->sig, &save->action, NULL) != 0) {
+			return -1;
+		}
+	}
+
+	return 0;
+}
